************************************************************************
      subroutine geomoptdrv(minpfile)
************************************************************************
* Driver for geometry optimization;
* parameters, specific arrays and types are declared in modul optim
************************************************************************
      use optim
      implicit none
      integer :: minpfile,natoms,natwdummy,npara,nopt
      integer :: nintcoord                              ! number of internal coordinates
      character(len=4) :: geom,unitc                    ! type of geometry (zmat/xyz) and its unit (Angstrom/Bohr)
      character(len=6) :: molden                       
      character(len=8) :: gopt,optalg,c8
      logical :: lmolden                                ! Is molden file requested?
      logical :: lcomp                                  ! whether the energy is composite or not

      double precision :: fgoal,optetol,optgtol,optstol
      double precision :: au                          ! scaling factor for coordinates au=1 for Bohr and au=ang2bohr for Angstrom
      integer :: optmaxit,maxfun,plevel,lnum
      integer :: istat
      logical :: bool                                 ! whether or not the optimization succeeded
      character(len=4)  :: c4
      character(len=16) :: c16



      integer i
      double precision, dimension(:,:),allocatable  :: Bmatrix   ! 
      double precision, dimension(:,:),allocatable  :: Pmat      ! 


#ifdef WINTEL
      logical :: fileexist
#endif
#if !defined (gfortran) && !defined (G95)
      integer*4 system
      external system
#endif

      interface

        subroutine read_MINP(minpfile,geom,
     $    gopt,npara,pname,fgopt,igopt,cgopt,nopt,
     $    natwdummy,natoms,atnums,au)
      implicit none
      integer,intent(in) :: minpfile
      character(len=4),intent(in) :: geom
      character(len=8),intent(in) :: gopt
      integer,intent(out) :: npara                        ! # parameters defined in the Z-matrix, or 3 x number of atoms in case of Cartesian cs
      integer,intent(out) :: nopt                         ! # parameters to be optimized
      character(len=*),dimension(:),intent(out) :: pname  ! name of the parameters
      double precision,dimension(:),intent(out) :: fgopt  ! value of parameters
      integer,dimension(:),intent(out)          :: igopt  ! whether or not the given parameter should be optimized
      character(len=2),dimension(:),intent(out) :: cgopt  ! to hold the special marks '++', '--', or '  '
      integer,intent(out) :: natwdummy,natoms             ! # atoms and # atoms including dummies
      integer,dimension(:),intent(out)          ::atnums  ! atomic numbers
      double precision,intent(in) :: au                   ! scaling factor for coordinates au=1 for Bohr and au=ang2bohr for Angstrom
        end subroutine read_MINP

        subroutine geomopt(minpfile,
     $    geom,npara,pname,fgopt,igopt,
     $    cgopt,nopt,optalg,   
     $    fgoal,ftol,gtol,xtol,maxiter,maxfun,plevel,
     $    Hessian,lmolden,bool,
     $    natwdummy,natoms,nintcoord,intcoords,rim,au,lcomp) !HB
      implicit none
      integer,intent(in) :: minpfile
      integer,intent(in) :: npara,nopt
      character(len=4),intent(in) :: geom
      logical,intent(in) :: lcomp !HB
      character(len=*),dimension(:),intent(in) :: pname  ! name of the parameters to be optimized
      double precision,dimension(:),intent(inout) :: fgopt  ! value of parameters to be optimized
      integer,dimension(:),intent(in)          :: igopt  ! whether or not the given parameter should be optimized
      character(len=2),dimension(:),intent(in) :: cgopt  ! to hold the special marks '++', '--', or '  '
      character(len=8),intent(in) :: optalg
      double precision,intent(in) :: fgoal
      double precision,intent(in) :: ftol 
      double precision,intent(in) :: gtol 
      double precision,intent(in) :: xtol
      integer,intent(in)          :: maxiter
      integer,intent(in)          :: maxfun
      integer,intent(in)          :: plevel
      double precision,dimension(:,:),intent(in) :: Hessian
      logical,intent(in)          :: lmolden            ! write molden file?
      logical,intent(inout)       :: bool               ! whether or not the optimization succeeded
      integer,intent(in)          :: natwdummy,natoms   ! # atoms and # atoms including dummies
      integer,intent(in)          :: nintcoord          ! # internal coordinates 
      double precision, dimension(nintcoord),intent(inout) :: intcoords ! csonti
      integer,dimension(5,nintcoord),intent(in) :: rim  ! red_int_matrix
      double precision,intent(in) :: au                 ! scaling factor for coordinates au=1 for Bohr and au=ang2bohr for Angstrom
        end subroutine geomopt

        subroutine MINP_zmat2cart(atsymbol,coordxyz,natwdummy,natoms)
      implicit none
      character(len=2),dimension(:),intent(in)   :: atsymbol
      double precision,dimension(:,:),intent(in) :: coordxyz
      integer,intent(out)                        :: natwdummy,natoms  ! # atoms and # atoms including dummies
        end subroutine MINP_zmat2cart
 
      end interface
     
! take the parameters from the MINP file
      call getkey('gopt',4,gopt,8)
      if(gopt.ne.'off     ') then
        write(*,*)'gopt is on: ',gopt
      else
        write(*,*)'gopt is off: ',gopt
        write(*,*)'you should not be here - csonti from geomopt'
      endif


      call getkey('verbosity',9,c4,4)
      read(c4,'(i4)') plevel
      call getkey('optalg',6,optalg,8)
      call getkey('optmaxit',8,c8,8)
      read(c8, '(i4)') optmaxit
      call getkey('oniom',5,c8,8)
      lcomp=.false.
      if(trim(c8).ne.'off') lcomp=.true.

      maxfun = optmaxit*15 ! this is quite arbitrary but should be fine in most situations

      call getkey('optetol',7,c16,16)
      read(c16, *) optetol
      call getkey('optgtol',7,c16,16)
      read(c16, *) optgtol
      call getkey('optstol',7,c16,16)
      read(c16, *) optstol

      lmolden = .False.
      call getkey('molden',6,molden,6)
      if(molden.eq.'on    ') lmolden=.True.
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

      istat=system('cp MINP MINP.init')
       if(istat.ne.0) then
         write(6,*) 'Error while copying MINP to MINP.init'
         stop
       endif

C Check geometry type and unit
      call getkey('geom',4,geom,4)
      if(geom.ne.'xyz '.and.
     $   geom.ne.'zmat'.and.
     $   geom.ne.'tmol'.and.
     $   geom.ne.'mol ') call unknown('geom',4)

      call getkey('unit',4,unitc,4) ! whether coordinates are in Angstrom or Bohr
      if(unitc .eq. 'bohr') then
        au = 1.d0 ! no scaling
      elseif(unitc .eq. 'angs') then
        au = ang2bohr
      endif

      call get_natoms(minpfile,geom,natwdummy,natoms) ! number of atoms

      if(geom.eq.'zmat') then ! convert it to Cartesian
        allocate(alloc_atsymbols(natwdummy)) !csonti allocate
        allocate(alloc_coords(3,natwdummy))

        allocate(alloc_zpname(3*natwdummy))          ! csonti allocate (upper limit), also introduce some error handling!!!
        allocate(alloc_zpvalue(3*natwdummy))         ! deallocate later!!!
        call cfillzero(alloc_zpname,3*natwdummy)
        call dfillzero(alloc_zpvalue,3*natwdummy)
!       call MINP_zmatvars(....

        open(minpfile,file='MINP')
        call getkeym('geom',4,c4,4)
        call zmattocart(alloc_coords,natwdummy,alloc_atsymbols) ! it gets the Cartesian-coordinates from the Z-matrix  
        close(minpfile)

        call MINP_zmat2cart(alloc_atsymbols,alloc_coords,natwdummy, ! it change the Z-matrix to Cartesian in the file MINP
     $                                                       natoms)
        if(allocated(alloc_atsymbols)) deallocate(alloc_atsymbols)
        if(allocated(alloc_coords)) deallocate(alloc_coords)
		
#ifdef WINTEL
        inquire(file="MINP.header", exist=fileexist)
        if(fileexist) then 
          inquire(file="MINP.footer", exist=fileexist)
          if(fileexist) then
#else
        istat=system('test -e MINP.header')
        if(istat.eq.0) then 
          istat=system('test -e MINP.footer')
          if(istat.eq.0) then
#endif
            istat=system('cat MINP.footer >> MINP.header')
            istat=system('mv MINP.header MINP')
            istat=system('rm -f MINP.footer') 
          endif
          istat=system('rm -f MINP.header') 
        endif
        geom='xyz '
      endif !end of conversion, MINP contains Cartesians and MINP.init contains zmat

#ifdef WINTEL
      inquire(file="DIIS.geo", exist=fileexist)
      if(fileexist) istat=system('rm -f DIIS.geo')
      inquire(file="DIISE.geo", exist=fileexist)
      if(fileexist) istat=system('rm -f DIISE.geo')
      inquire(file="DIISG.geo", exist=fileexist)
      if(fileexist) istat=system('rm -f DIISG.geo')
#else
      istat=system('test -e DIIS.geo')
      if(istat.eq.0) istat=system('rm -f DIIS.geo')
      istat=system('test -e DIISE.geo')
      if(istat.eq.0) istat=system('rm -f DIISE.geo')
      istat=system('test -e DIISG.geo')
      if(istat.eq.0) istat=system('rm -f DIISG.geo')
#endif

      allocate(alloc_atnums(natoms))                   ! csonti allocate insert error handling
      allocate(alloc_pname(natoms))                    ! csonti allocate
      allocate(alloc_fgopt(3*natoms))                    ! csonti allocate
      allocate(alloc_igopt(3*natoms))                    ! csonti allocate
      allocate(alloc_cgopt(3*natoms))                    ! csonti allocate

      call read_MINP(minpfile,geom,gopt,npara,
     $  alloc_pname,alloc_fgopt,alloc_igopt,alloc_cgopt,
     $  nopt,natwdummy,natoms,alloc_atnums,au)
!     write(*,*)'nopt: ', nopt

      allocate(alloc_coords(3,natoms))                              ! allocate array for the Cartesian coordinates
      alloc_coords = get_cartcoords(natoms,alloc_fgopt)

!     write(*,*) alloc_atnums
      call build_red_int_matrix(alloc_atnums,alloc_coords,alloc_rim)! redundant internal matrix (see module optim for details)
      
      nintcoord = size(alloc_rim,2)
      allocate(alloc_intcoords(nintcoord))                          ! allocate array for the Cartesian coordinates

      alloc_intcoords = get_intcoords(natoms,alloc_coords,nintcoord,
     $                                alloc_rim)
      call wrt_intcoords(nintcoord,alloc_intcoords)
!     call read_intcoords(nintcoord,intcoords)

      allocate(alloc_hessian(nintcoord,nintcoord)) ! csonti allocate initial Hessian matrix
!     call build_unit_Hessian(alloc_hessian)
!     call build_initial_Hessian(alloc_hessian,alloc_rim,alloc_coords)
      call build_Lindh_Hessian(alloc_hessian,alloc_rim,
     $                         alloc_coords,alloc_atnums)
!     call build_Schlegel_Hessian(alloc_hessian,alloc_rim,
!    $                         alloc_coords,alloc_atnums)
!     write(*,*) 'Initial Hessian matrix:'
!     call print_matrixR(alloc_hessian,10)

      allocate(Bmatrix(nintcoord,3*natoms))
      allocate(Pmat(nintcoord,nintcoord))

      call build_Wilson_Bmat(natoms,alloc_coords,alloc_rim,Bmatrix)     ! it builds Wilson's B-matrix
!     write(*,*) 'Wilsons B matrix:'
!     call print_matrixR(Bmatrix,10)

      call build_projection(Bmatrix,Pmat)
!     write(*,*) 'Projection matrix:'
!     call print_matrixR(Pmat,10)

      alloc_hessian=matmul(alloc_hessian,Pmat)
      alloc_hessian=matmul(Pmat,alloc_hessian)
!     write(*,*) 'Projected Hessian matrix:'
!     call print_matrixR(alloc_hessian,10)

!     call matrix_pseudoinverse(alloc_hessian,Pmat)
!     alloc_hessian = Pmat

      bool=.False.
!     if(optalg.eq.'simplex ') then
        call geomopt(minpfile,geom,npara,
     $    alloc_pname,alloc_fgopt,alloc_igopt,
     $    alloc_cgopt,nopt,optalg,   
     $    fgoal,optetol,optgtol,optstol,optmaxit,maxfun,plevel,
     $    alloc_hessian,lmolden,bool,
     $    natwdummy,natoms,
     $    nintcoord,alloc_intcoords,alloc_rim,au,lcomp)
!     else
!       write(*,*) 'At the moment this should not happen 
!    $- csonti from geomopt.f'
!
!     endif

      if(bool) then
         istat=system('mv MINP.tmp MINP.opt')
         if(istat.ne.0) then
           write(6,*) 'Error while renaming MINP.tmp to MINP.opt'
           stop
         endif
      endif

!     remove temp files
#ifdef WINTEL
      inquire(file="MINP.header", exist=fileexist)
      if(fileexist) istat=system('rm -f MINP.header') 
      inquire(file="MINP.footer", exist=fileexist)
      if(fileexist) istat=system('rm -f MINP.footer') 
#else
      istat=system('test -e MINP.header')
      if(istat.eq.0) istat=system('rm -f MINP.header') 
      istat=system('test -e MINP.footer')
      if(istat.eq.0) istat=system('rm -f MINP.footer')
#endif

      if(allocated(alloc_hessian)) deallocate(alloc_hessian)  !csonti deallocate

      return
      end subroutine geomoptdrv
************************************************************************

************************************************************************
      subroutine get_natoms(minpfile,geom,natwdummy,natoms)
************************************************************************
* It returns the number of atoms including dummies (natwdummy)
* and the number of atoms without dummies (natoms)
************************************************************************
      implicit none
      integer,intent(in)          :: minpfile
      character(len=4),intent(in) :: geom
      integer,intent(out)         :: natwdummy,natoms ! # atoms and # atoms including dummies
      !local
      integer :: check
      character(len=512) :: line,stmp
      logical :: lgeom = .False.
      character(len=10) :: str10

      natoms=0
      natwdummy=0
      check = 0

      line=''
      stmp=adjustl(trim(line))

      if(geom.eq.'zmat') then
        call ZMAT()
      elseif(geom.eq.'xyz ') then
        call Cartesian()
      else
       write(*,*) 'Not supported geometry specification'
       write(*,*) 'Only Z-matrix and Cartesian coordinates are allowed.'
       call mrccend(1)
      endif

!     ******************************************************************
      contains
!     ******************************************************************

!     ******************************************************************
      subroutine ZMAT
!     ******************************************************************

      open(minpfile,file='MINP',status='old')

      do while (check.eq.0 .and. stmp(1:4).ne.'geom')
        read(minpfile,200,iostat=check) line        ! lines before keyword geom
        stmp=adjustl(trim(line))
        if(stmp(1:4).eq.'geom') lgeom=.True.
      enddo

      if(lgeom) then
        do while (check.eq.0 .and. trim(line).ne.'')     ! Z-matrix definition part of geometry specification
          read(minpfile,200,iostat=check) line
          stmp = adjustl(trim(line))
          call lowercase(stmp,stmp,2)
          natwdummy=natwdummy+1
          if(stmp(1:1).ne.'x' .or. stmp(2:2).eq.'e') natoms=natoms+1
          if(stmp(1:1).eq.' ') then
            natoms=natoms-1
            natwdummy=natwdummy-1
          endif
        enddo

      endif ! lgeom

      close(minpfile)

200   format(a512)
      end subroutine ZMAT
!     ******************************************************************

!     ******************************************************************
      subroutine Cartesian()

      use optim, only: integer_, str_cut
      implicit none
      !local
      character(len=512) :: stra
      character(len=512) :: strsy                ! temporary string for atom's symbol or number
      integer :: i

      line=''
      stmp=adjustl(trim(line))

      open(minpfile,file='MINP',status='old')

      do while (check.eq.0 .and. stmp(1:4).ne.'geom')
        read(minpfile,200,iostat=check) line        ! lines before keyword geom
        stmp=adjustl(trim(line))
        if(stmp(1:4).eq.'geom') lgeom=.True.
      enddo

      if(lgeom) then
        read(minpfile,200,iostat=check) line          ! number of atoms
        stmp=adjustl(trim(line))
        natwdummy = integer_(stmp)

        read(minpfile,200,iostat=check) line          ! title line of geom. spec.
        stmp=adjustl(trim(line))

        do i=1,natwdummy                                 ! read Cartesian coordinates including dummies
          read(minpfile,200,iostat=check) line
          line = adjustl(trim(line))
          call str_cut(line,' ',strsy,stra)
          stmp = adjustl(trim(strsy))             ! atomic symbol/number (dummy symbol X handled - no atomic number for dummies!!!)
          call lowercase(stmp,stmp,2)
          if(stmp(1:1).ne.'x' .or. stmp(2:2).eq.'e') natoms=natoms+1

        enddo
      endif ! lgeom

      close(minpfile)

200   format(a512)
      end subroutine Cartesian
!     ******************************************************************

      end subroutine get_natoms
************************************************************************

************************************************************************
      subroutine MINP_zmatvars(atsymbol,coordxyz,zpname,zpvalue)
************************************************************************
* It determines the variables defined in the Z-matrix 
************************************************************************
      use optim, only: double_, str_cut, cfillzero
      implicit none
      double precision,dimension(:,:),intent(in) :: coordxyz
      character(len=2),dimension(:),intent(in)   :: atsymbol
      character(len=*),dimension(:),intent(out)  :: zpname
      double precision,dimension(:),intent(out)  :: zpvalue
      !local
      integer :: check
      integer,parameter :: minpfile=68,mheaderf=80,mfooterf=81
      character(len=512) :: line,stmp
      character(len=512) :: stra,strb
      integer :: i,j,l
      logical :: lgeom = .False.
      integer :: natwdummy,natoms              ! # atoms and # atoms including dummies
      integer :: nzp                           ! # of Z-matrix parameters
      character(len=10) :: str10


      natoms=0
      natwdummy=0
      nzp = 0
      check = 0

      line=''
      stmp=adjustl(trim(line))

      open(minpfile,file='MINP',status='old')
      open(mheaderf,file='MINP.header',status='unknown')
      open(mfooterf,file='MINP.footer',status='unknown')

      do while (check.eq.0 .and. stmp(1:4).ne.'geom')
        read(minpfile,200,iostat=check) line        ! lines before keyword geom
        stmp=adjustl(trim(line))
        l = len_trim(stmp)
        if(stmp(1:4).eq.'geom') then 
          lgeom=.True.
        else
          write(mheaderf,'(a)') stmp(1:l)
        endif
      enddo

      if(lgeom) then
        do while (check.eq.0 .and. trim(line).ne.'')     ! Z-matrix definition part of geometry specification
          read(minpfile,200,iostat=check) line
          stmp = adjustl(trim(line))
          call lowercase(stmp,stmp,2)
          natwdummy=natwdummy+1
          if(stmp(1:1).ne.'x' .or. stmp(2:2).eq.'e') natoms=natoms+1
          if(stmp(1:1).eq.' ') then
            natoms=natoms-1
            natwdummy=natwdummy-1
          endif
        enddo

        line='%' ! It is just a trick so that the routine continues file reading

        do while (check.eq.0 .and. trim(line).ne.'')     ! definition of Z-matrix variables
          read(minpfile,200,iostat=check) line
            if(trim(line).ne.'') then
              nzp = nzp+1 ! number of parameters in the Z-matrix
              call str_cut(line,'=',stra,strb)
              zpname(nzp) = stra ! name of the parameter in the Z-matrix
!             call str_dtest(strb,fgopt(npara),lopt,mark)
              zpvalue(nzp) = double_(strb)
            endif
        enddo
      endif ! lgeom


        do while (check.eq.0 .and. trim(line).ne.'')     ! definition of Z-matrix variables
          read(minpfile,200,iostat=check) line
        enddo



      write(mheaderf,'(a)')   'geom=xyz'                 ! write Cartesian coords
      write(str10,'(i10)')    natoms
      write(mheaderf,'(a)')   adjustl(str10)
      write(mheaderf,'(a)')   ''

      do i=1,natoms
        write(mheaderf,"(a2,3f20.10)")
     $        adjustl(atsymbol(i)),(coordxyz(j,i),j=1,3)
      enddo
      write(mheaderf,'(a)')   ''

      do while (check.eq.0)     ! lines after geometry specification
          line=''
          read(minpfile,200,iostat=check) line
          if(check .eq. 0) then
            stmp = adjustl(trim(line))
            l = len_trim(stmp)
            write(mfooterf,'(a)') stmp(1:l)
          endif
      enddo


      close(mheaderf)
      close(mfooterf)
      close(minpfile)

200   format(a512)

      end subroutine MINP_zmatvars
************************************************************************

************************************************************************
      subroutine MINP_zmat2cart(atsymbol,coordxyz,natwdummy,natoms)
************************************************************************
* It converts MINP with geom. spec. in ZMAT 
* to 
* MINP with geom. spec. in Cartesian
************************************************************************
      use optim, only: double_, str_cut, cfillzero
      implicit none
      double precision,dimension(:,:),intent(in) :: coordxyz
      character(len=2),dimension(:),intent(in)   :: atsymbol
      integer,intent(in)                         :: natwdummy,natoms  ! # atoms and # atoms including dummies
      !local
      integer :: check
      integer,parameter :: minpfile=68,mheaderf=80,mfooterf=81
      character(len=512) :: line,stmp
      character(len=512) :: stra,strb
      integer :: i,j,l
      logical :: lgeom = .False.
      character(len=10) :: str10


      check = 0

      line=''
      stmp=adjustl(trim(line))

      open(minpfile,file='MINP',status='old')
      open(mheaderf,file='MINP.header',status='unknown')
      open(mfooterf,file='MINP.footer',status='unknown')

      do while (check.eq.0 .and. stmp(1:4).ne.'geom')
        read(minpfile,200,iostat=check) line        ! lines before keyword geom
        stmp=adjustl(trim(line))
        l = len_trim(stmp)
        if(stmp(1:4).eq.'geom') then 
          lgeom=.True.
        else
          write(mheaderf,'(a)') stmp(1:l)
        endif
      enddo

      if(lgeom) then
        do while (check.eq.0 .and. trim(line).ne.'')     ! Z-matrix definition part of geometry specification
          read(minpfile,200,iostat=check) line
        enddo

        line='%' ! It is just a trick so that the routine continues file reading

        do while (check.eq.0 .and. trim(line).ne.'')     ! definition of Z-matrix variables
          read(minpfile,200,iostat=check) line
        enddo
      endif ! lgeom


      write(mheaderf,'(a)')   'geom=xyz'                 ! write Cartesian coords
      write(str10,'(i10)')    natoms
      write(mheaderf,'(a)')   adjustl(str10)
      write(mheaderf,'(a)')   ''

      do i=1,natoms
        write(mheaderf,"(a2,3f20.10)")
     $        adjustl(atsymbol(i)),(coordxyz(j,i),j=1,3)
      enddo
      write(mheaderf,'(a)')   ''

      do while (check.eq.0)     ! lines after geometry specification
          line=''
          read(minpfile,200,iostat=check) line
          if(check .eq. 0) then
            stmp = adjustl(trim(line))
            l = len_trim(stmp)
            write(mfooterf,'(a)') stmp(1:l)
          endif
      enddo


      close(mheaderf)
      close(mfooterf)
      close(minpfile)

200   format(a512)

      end subroutine MINP_zmat2cart
************************************************************************


************************************************************************
      subroutine read_MINP(minpfile,geom,
     $   gopt,npara,pname,fgopt,igopt,cgopt,nopt,
     $   natwdummy,natoms,atnums,au)

************************************************************************
*
************************************************************************
      use optim
      implicit none
      integer,intent(in) :: minpfile
      character(len=4),intent(in) :: geom
      character(len=8),intent(in) :: gopt
      integer,intent(out) :: npara                         ! # parameters defined in the Z-matrix, or 3*natoms in Cartesian
      integer,intent(out) :: nopt                          ! # parameters to be optimized
      character(len=*),dimension(:),intent(out)  :: pname  ! name of the parameters
      double precision,dimension(:),intent(out)  :: fgopt  ! value of parameters
      integer,dimension(:),intent(out)           :: igopt  ! whether or not the given parameter should be optimized
      character(len=2),dimension(:),intent(out)  :: cgopt  ! to hold the special marks '++', '--', or '  '
      integer,intent(out) :: natwdummy,natoms              ! # atoms and # atoms including dummies
      integer,dimension(:),intent(out)           ::atnums  ! atomic numbers
      double precision,intent(in) :: au                    ! scaling factor for coordinates au=1 for Bohr and au=ang2bohr for Angstrom
      !locals
      integer :: check
      character(len=512) :: line,stmp
      character(len=2) :: mark = '  '
      logical :: lgeom = .False.,lopt=.True.

      natoms=0
      natwdummy=0
      npara = 0
      check = 0

      if(geom.eq.'zmat') then
        call ZMAT()
      elseif(geom.eq.'xyz ') then
        call Cartesian()
      else
       write(*,*) 'Not supported geometry specification'
       write(*,*) 'Only Z-matrix and Cartesian coordinates are allowed.'
       call mrccend(1)
      endif


!     ******************************************************************
      contains                                             ! read_MINP
!     ******************************************************************

!     ******************************************************************
      subroutine ZMAT()
      implicit none
      !local
      character(len=512) :: stra,strb
      integer :: i,l
      integer,parameter :: nparamax=150 ! csonti remove constraint 

      line=''
      stmp=adjustl(trim(line))

      open(minpfile,file='MINP',status='old')
      open(mheaderf,file='MINP.header',status='unknown')
      open(mfooterf,file='MINP.footer',status='unknown')

      do while (check.eq.0 .and. stmp(1:4).ne.'geom')
        read(minpfile,200,iostat=check) line        ! lines before keyword geom
        stmp=adjustl(trim(line))
        l = len_trim(stmp)
        write(mheaderf,'(a)') stmp(1:l)
        if(stmp(1:4).eq.'geom') lgeom=.True.
      enddo

      if(lgeom) then
        do while (check.eq.0 .and. trim(line).ne.'')     ! Z-matrix definition part of geometry specification
          read(minpfile,200,iostat=check) line
          stmp = adjustl(trim(line))
          call lowercase(stmp,stmp,2)
          natwdummy=natwdummy+1
          if(stmp(1:1).ne.'x' .or. stmp(2:2).eq.'e') natoms=natoms+1
        enddo

        line='%' ! It is just a trick so that the routine continues file reading

        nopt=0
        do while (check.eq.0 .and. trim(line).ne.'')     ! definition of Z-matrix variables
          read(minpfile,200,iostat=check) line
          if(trim(line).ne.'') then
            npara=npara+1
            call str_cut(line,'=',stra,strb)
            pname(npara) = stra
            if(gopt.eq.'freeze  ') then   ! global freeze
              lopt = .False.
            else
              lopt = .True.
            endif
            mark = '  '
            call str_dtest(strb,fgopt(npara),lopt,mark)
            cgopt(npara)=mark
            if(.not.lopt) then
              igopt(npara)=0
            else
              igopt(npara)=1
              nopt=nopt+1
            endif
          endif
        enddo
      endif ! lgeom

      do while (check.eq.0)     ! lines after geometry specification
          line=''
          read(minpfile,200,iostat=check) line
          if(check .eq. 0) then
            stmp = adjustl(trim(line))
            l = len_trim(stmp)
            write(mfooterf,'(a)') stmp(1:l)
          endif
      enddo


      close(mheaderf)
      close(mfooterf)
      close(minpfile)

200   format(a512)
      write(*,*) 'this is the end!'
      call mrccend(1)

      end subroutine ZMAT
!     ******************************************************************

!     ******************************************************************
      subroutine Cartesian()
      implicit none
      !local
      character(len=512) :: stra,strb
      character(len=512) :: strsy                ! temporary string for atom's symbol or number
      character(len=512),dimension(3) :: scoord  ! temporary strings for the atom's x,y,z-coords
      integer :: i,j,k,l

      line=''
      stmp=adjustl(trim(line))

      open(minpfile,file='MINP',status='old')
      open(mheaderf,file='MINP.header',status='unknown')
      open(mfooterf,file='MINP.footer',status='unknown')

      do while (check.eq.0 .and. stmp(1:4).ne.'geom')
        read(minpfile,200,iostat=check) line        ! lines before keyword geom
        stmp=adjustl(trim(line))
        l = len_trim(stmp)
        write(mheaderf,'(a)') stmp(1:l)
        if(stmp(1:4).eq.'geom') lgeom=.True.
      enddo

      if(lgeom) then
        read(minpfile,200,iostat=check) line          ! number of atoms
        stmp=adjustl(trim(line))
        l = len_trim(stmp)
        write(mheaderf,'(a)') stmp(1:l)
        natwdummy = integer_(stmp)


        read(minpfile,200,iostat=check) line          ! title line of geom. spec.
        stmp=adjustl(trim(line))
        l = len_trim(stmp)
        write(mheaderf,'(a)') stmp(1:l)

        nopt=0
        do i=1,natwdummy                                 ! read Cartesian coordinates including dummies
          read(minpfile,200,iostat=check) line
          line = adjustl(trim(line))
          call str_cut(line,' ',strsy,stra)
          pname(i) = adjustl(trim(strsy))             ! atomic symbol/number (dummy symbol X handled - no atomic number for dummies!!!)
          stmp = pname(i)
          atnums(i) = get_atomic_number(pname(i))
          call lowercase(stmp,stmp,2)
          if(stmp(1:1).ne.'x' .or. stmp(2:2).eq.'e') natoms=natoms+1

          stra = adjustl(trim(stra))
          call str_cut(stra,' ',scoord(1),strb)
          strb = adjustl(trim(strb))
          call str_cut(strb,' ',scoord(2),stra)
          stra = adjustl(trim(stra))
          call str_cut(stra,' ',scoord(3),strb)

          do j=1,3
            if(gopt.eq.'freeze  ') then   ! global freeze
              lopt = .False.
            else
              lopt = .True.
            endif
            mark = '  '
            call str_dtest(scoord(j),fgopt(npara+j),lopt,mark)
            fgopt(npara+j) = fgopt(npara+j)*au !scale to Bohr
            cgopt(npara+j)=mark
            if(.not.lopt) then
              igopt(npara+j)=0
            else
              igopt(npara+j)=1
              nopt=nopt+1
            endif
          enddo !j
          npara=npara+3

        enddo
      endif ! lgeom

      do while (check.eq.0)     ! lines after geometry specification
          line=''
          read(minpfile,200,iostat=check) line
          if(check .eq.0) then
            stmp = adjustl(trim(line))
            l = len_trim(stmp)
            write(mfooterf,'(a)') stmp(1:l)
          endif
      enddo


      close(mheaderf)
      close(mfooterf)
      close(minpfile)

200   format(a512)


      end subroutine Cartesian
!     ******************************************************************

      end subroutine read_MINP
************************************************************************




************************************************************************
      subroutine geomopt(minpfile,geom,
     $    npara,pname,fgopt,igopt,cgopt,nopt,
     $    optalg,
     $    fgoal,ftol,gtol,xtol,maxiter,maxfun,plevel,
     $    Hessian,lmolden,bool,
     $    natwdummy,natoms,nintcoord,intcoords,rim,au,lcomp)
************************************************************************
      use optim
      implicit none
      integer,intent(in) :: minpfile
      integer,intent(in) :: npara,nopt
      character(len=4),intent(in) :: geom
      character(len=*),dimension(:),intent(in):: pname     ! name of the parameters to be optimized
      double precision,dimension(:),intent(inout) :: fgopt  ! parameter array to be optimized, Cartesian coordinates are in row-major order (x1,y1,z1,x2,y2,..)
      integer,dimension(:),intent(in)          :: igopt     ! whether or not the given parameter should be optimized
      character(len=2),dimension(:),intent(in) :: cgopt     ! to hold the special marks '++', '--', or '  '
      character(len=8),intent(in) :: optalg
      double precision,intent(in) :: fgoal
! ***
! simplex parameters     
! ftol    ->  convergence criterion for the function values
! gtol    ->  convergence criterion for the function values
! xtol    ->  convergence criterion for the function arguments
! maxiter ->  maximum number of iterations allowed
! maxfun  ->  maximum number of function evaluations allowed
! plevel  ->  determine the amount of information to be printed (the larger the more)
! bool    ->  whether or not the optimization succeeded
! finit   ->  initial value of f2min
! fopt    ->  minimal value of f2min
! iter    ->  number of iterations performed
! nfeval  ->  number of function evaluations performed
! ***
      double precision,intent(in) :: ftol
      double precision,intent(in) :: gtol
      double precision,intent(in) :: xtol
      integer,intent(in)          :: maxiter
      integer,intent(in)          :: maxfun
      integer,intent(in)          :: plevel
      double precision,dimension(:,:),intent(in) :: Hessian
      logical,intent(in)          :: lmolden            ! write molden file?
      logical,intent(inout)       :: bool               ! whether or not the optimization succeeded
! end simplex parameters

      integer,intent(in)          :: natwdummy,natoms   ! # atoms and # atoms including dummies
      integer,intent(in)          :: nintcoord          ! # internal coordinates 
      double precision, dimension(nintcoord),intent(inout) :: intcoords ! csonti
      integer,dimension(5,nintcoord),intent(in) :: rim  ! red_int_matrix
      double precision,intent(in) :: au                 ! scaling factor for coordinates au=1 for Bohr and au=ang2bohr for Angstrom


      !local
      integer :: ipara,ndim,iter,nfeval,ngeval,i
      double precision para(nopt),finit,fopt,gradient(nopt)
      logical, parameter :: logopt=.False. ! whether the parameter space's logarithm is taken or not
      logical, dimension(nopt) :: lns     ! to take notes of the parameter's sign before log()
      logical lcomp
!

c#if !defined (gfortran) && !defined (G95)
c      integer*4 system
c      external system
c#endif

! initial guess for the parameters to be optimized
      
      ipara=0
      do i=1,npara
        if(igopt(i).gt.0) then
          ipara=ipara+1
          para(ipara)=fgopt(i)
        endif
      enddo

! end initial guess for the parameters to be optimized

! taking the logarithm of the arguments if the optimization takes place 
! in that space
!     if (logopt) call log_trans(para,lns)

!***********************************************************************

      write(*,1003) 
      write(*,'(21x,a)') 'Geometry Optimization has started'
      write(*,*) '' 

      if(optalg.eq.'simplex ') then
        finit = goalfunc(para)! initial function value csonti include in simplex as a return value like in bfgs!!!

        call simplex(goalfunc,para,ftol,xtol,maxiter,maxfun,
     $             fopt,iter,nfeval,plevel,bool)
      elseif(optalg.eq.'bfgs    ') then
!       call bfgs(mrccEGrad,intcoords,Hessian,
        call bfgs_B(mrccEGrad,intcoords,Hessian,
     $             ftol,gtol,xtol,maxiter,maxfun,
     $             finit,fopt,iter,nfeval,ngeval,plevel,lmolden,bool)
      else
        write(*,*) 'This is impossible ... (csonti)'
      endif

      write(*,1003) 
      write(*,*) '' 
      write(*,'(a40,f17.8)') 'Initial function value:',finit
      write(*,*) '' 
      write(*,'(a40,f17.8)') 'Optimal function value:',fopt
      write(*,'(a42,d16.3)') 'Energy tolerance:',ftol
      write(*,'(a42,d16.3)') 'Gradient tolerance:',gtol
      write(*,'(a42,d16.3)') 'Step tolerance:',xtol
      write(*,'(a42,i8)') 'Number of iterations:',iter
      write(*,'(a42,i8)') 'Number of function evaluations:',nfeval
      write(*,'(a42,i8)') 'Number of gradient evaluations:',nfeval
      write(*,*) '' 
      if (bool) then
      write(*,'(15x,a)')'Geometry Optimization has ended successfully'
      else
      write(*,'(21x,a)')'Geometry Optimization has failed'
      endif
      write(*,1003) 
1003  format(1x,70('%'))

      return

!     ******************************************************************
      contains                                            !geomopt
!     ******************************************************************

!     ******************************************************************
      function mrccEGrad(para)
!     ******************************************************************
!     * it executes MRCC and returns both the gradient and the energy
!     ******************************************************************
      implicit none
      double precision, dimension(:), intent(inout) :: para
      double precision, dimension(size(para)+1) :: mrccEGrad ! the last element is the energy!

      !local
      integer i,j,ndim

      ! coordinate transformation matrix and its inverse
      double precision, dimension(3,3) :: transf_m, transf_m_inv
      double precision, dimension(3)   :: centmass ! center of mass
      real*8 worka(3)
      real*8 gradInStd(3,natoms),gradInStdT(natoms,3)
      real*8 gradtmp(3,natoms),  gradtmpT(natoms,3)
      integer info,ipiv(3),gradfile

      double precision, dimension(size(para)):: grads
      double precision, dimension(3,natoms)  :: coords,coords0,coords1
      double precision, dimension(3*natoms)  :: x0,xk,gradvec
      double precision, dimension(nintcoord) :: intcoords0,intcoords1,sq! intcoord0 -> previous values from file COORD.int, intcoord1 -> current values

      double precision, dimension(nintcoord,3*natoms)  :: Bmatrix   ! 
      double precision, dimension(3*natoms,nintcoord)  :: Bt        ! 
      double precision, dimension(nintcoord,3*natoms)  :: Btp       ! 
      double precision, dimension(nintcoord,nintcoord) :: Pmat      ! 

!     ******************************************************************

      ndim = size(para)
      do i=1,ndim+1                                          ! +1 because of energy
        mrccEGrad(i) = 0.d0
      enddo

!     write(*,'(A20,99f15.6)')'para: ', para
!     write(*,'(A20,99f15.6)')'intc: ', intcoords

      call read_intcoords(nintcoord,intcoords0)              ! it gets the previous coordinates /intcoords0/
      coords0    = get_cartcoords(natoms,fgopt)              ! it extracts the Cartesian coords from fgopt
!     intcoords1 = para                                      ! current coordinates
      do i=1,nintcoord
        if(rim(1,i) .eq. 2 .or. rim(1,i) .eq. 3) then        ! it handles the periodicity of angles and dihedrals
          intcoords1(i) = mod(para(i),2.0d0*PI)
        else
          intcoords1(i) = para(i)
        endif
      enddo

      sq = intcoords1 - intcoords0                           ! difference

!     write(*,'(A20,99f15.6)')'intc0: ', intcoords0
!     write(*,'(A20,99f15.6)')'intc1: ', intcoords1
!     write(*,'(A20,99f15.6)') 'sq: ',sq

      call build_Wilson_Bmat(natoms,coords0,rim,Bmatrix)     ! it builds Wilson's B-matrix
!     write(*,*) 'Wilson''s B-matrix :'
!     call print_matrixR(Bmatrix,10)

      call matrix_transpose(Bmatrix,Bt)
!     write(*,*) 'Transpose of Wilson''s B-matrix (B^T):'
!     call print_matrixR(Bt,10)

      call matrix_pseudoinverse(Bt,Btp)
!     write(*,*) 'Pseudo-inverse of B^T:'
!     call print_matrixR(Btp,10)

      call build_projection(Bmatrix,Pmat)
!     write(*,*) 'Projection matrix:'
!     call print_matrixR(Pmat,10)
      call wrt_Pmat(nintcoord,Pmat)
!
      sq = matrix_dot_vec(Pmat,sq)
      x0=reshape(coords0,shape(x0))                          ! it flattens coords0; 3 x natoms -> 1 x 3N

      call step_transform(x0,intcoords0,Bmatrix,sq,xk,
     $                    natoms,nintcoord,rim)
!     fgopt(1:npara) = xk
      fgopt = xk

!     write(*,*) 'x0: '
!     call print_vectorR(x0,10)
!     write(*,*) 'xk: '
!     call print_vectorR(xk,10)


!     call int2cart(para,coords)

      if(geom.eq.'zmat') then
        call update_MINP_ZMAT(para)                          ! csonti attention
      elseif(geom.eq.'xyz ') then
       call update_MINP_Cart(xk,au)
       call wrt_intcoords(nintcoord,intcoords1)
      else
       write(*,*) 'Not supported geometry specification'
       write(*,*) 'Only Z-matrix and Cartesian coordinates are allowed.'
       call mrccend(1)
      endif

! execute MRCC and extract energy
      if(lcomp) then !HB
       call composite(.true.)
      else
       call spoint(.true.)
      endif
      mrccEGrad(ndim+1) = get_last_energy()

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

! transform MRCC gradient

      call getvar('centmass  ',centmass)
!     write(*,*)'centmass:',centmass
      call getvar('ten       ',transf_m)

!     call print_matrixR(transf_m,10)

!     transf_m_inv=transpose(transf_m)
!     transf_m_inv = transf_m
!     call dgetrf(3,3,transf_m_inv,3,ipiv,info)
!     call dgetri(3,transf_m_inv,3,ipiv,worka,3,info)

      gradfile=99
      open(gradfile,file='GRAD',status='old',form='unformatted')
        do i=1,natoms
         read(gradfile) 
     $          gradInStdT(i,1),gradInStdT(i,2),gradInStdT(i,3)
        enddo
      close(gradfile)

      gradInStd = transpose(gradInStdT)
      call dgemm('t','n',3,natoms,3,1.d0,transf_m,3,gradInStd,3,
     $            0.d0,gradtmp,3)


      gradvec=reshape(gradtmp,shape(gradvec))

!     call read_Pmat(nintcoord,Pmat)
!     write(*,*) 'Projection matrix after i/o:'
!     call print_matrixR(Pmat,10)

      grads = matrix_dot_vec(Btp,gradvec)
!     write(*,*)'Internal gradient vector:'
!     call print_vectorR(grads,5)

      grads = matrix_dot_vec(Pmat,grads)
!     write(*,*)'Internal gradient vector (projected):'
!     call print_vectorR(grads,5)

      mrccEGrad(1:ndim) = grads


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

!     grads = fnumgrad(para)
!     ndim = size(para)
!     do i=1,ndim                                          ! +1 because of energy
!       mrccEGrad(i) = grads(i)
!     enddo

      return

      end function mrccEGrad
!     ******************************************************************


!     ******************************************************************
      double precision function mrccE(para) ! this will be removed/or not csonti it depends on the implementation of numerical grads
!     ******************************************************************
!     * it executes MRCC and returns an energy, 
!     * and it also updates the MINP file
!     ******************************************************************
      implicit none
      double precision, dimension(:), intent(inout) :: para

      if(geom.eq.'zmat') then
        call update_MINP_ZMAT(para)
      elseif(geom.eq.'xyz ') then
        call update_MINP_Cart(para,au)
      else
       write(*,*) 'Not supported geometry specification'
       write(*,*) 'Only Z-matrix and Cartesian coordinates are allowed.'
       call mrccend(1)
      endif

! execute MRCC and extract energy
      if(lcomp) then !HB
       call composite(.true.)
      else
       call spoint(.true.)
      endif
      mrccE = get_last_energy()

      return

      end function mrccE
!     ******************************************************************

!     ******************************************************************
      double precision function goalfunc(para)
!     ******************************************************************
      implicit none
      double precision, dimension(:), intent(inout) :: para
      double precision :: res
      
      res = mrccE(para)
       if(dabs(fgoal).lt.1d-10) then
        goalfunc = res
       else
        goalfunc = dabs(fgoal-res)
       endif
      return
      end function goalfunc
!     ******************************************************************


!     ******************************************************************
      function fnumgrad(args)
!     ******************************************************************
!     * a simple two-point/central difference numerical gradient code 
!     * it calls function mrccE
!     ******************************************************************
      use optim
      implicit none
      double precision, dimension(:), intent(inout) :: args
      double precision, dimension(size(args)) :: fnumgrad
      !local
      integer :: i,j,ndim,gradfile
      double precision, dimension(size(args)) :: e, eb
      double precision :: eps,x0

      eps = sqrt ( epsilon_() ) ! square-root of machine epsilon
      ndim = size(args)
      do i=1,ndim
        fnumgrad(i) = 0.d0
      enddo

      x0 = mrccE(args)

      do i=1,ndim
        do j=1,ndim
          if( j .eq. i ) then
            e(j) = args(j)+eps
            eb(j) = args(j)-eps
          else
            e(j) = args(j)
            eb(j) = args(j)
          endif
        enddo
        fnumgrad(i) = (mrccE(e) - x0 ) / eps
!       fnumgrad(i) = (mrccE(e) - mrccE(eb)) / (2*eps)
      enddo

      gradfile=99

      open(gradfile,file='GRAD',status='unknown',form='unformatted')
      do i=1,ndim
        write(gradfile) fnumgrad(i)
      enddo
      close(gradfile)

! just a test      
      open(gradfile,file='GRAD',status='old',form='unformatted')
        read(gradfile) (fnumgrad(i), i=1,ndim)
      close(gradfile)

      return
      end function fnumgrad
!     ******************************************************************

!     ******************************************************************
      subroutine update_MINP_ZMAT(para)
!     ******************************************************************
!     * it updates the MINP file (Z-matrix version)
!     ******************************************************************
      implicit none

      double precision, dimension(:), intent(inout) :: para
      integer :: ipara,check,i,l
      character(len=512) :: line,stmp

!     do i=1,nopt ! test the parameters to be optimized
!       write(*,*) i, para(i)
!     enddo

      ipara=0
      do i=1,npara ! total parameters in Z-mat
        if(igopt(i).gt.0) then
          ipara=ipara+1
          fgopt(i)=para(ipara)
        endif
      enddo


      open(minpfile,file='MINP',status='old')
      open(minptmpf,file='MINP.tmp',status='unknown')
      open(mheaderf,file='MINP.header',status='old')
      open(mfooterf,file='MINP.footer',status='old')

      check = 0                                       ! MINP lines before geom spec
      do while (check.eq.0)
          line = ''
          read(mheaderf,204,iostat=check) line
          stmp = adjustl(trim(line))
          l = len_trim(stmp)
          if(check.eq.0) then
            write(minpfile,'(a)') stmp(1:l)
            write(minptmpf,'(a)') stmp(1:l)
          endif
      enddo

!     do i=1,ilb                                         ! lines before geom spec
!       write(minpfile,'(a)') adjustl(trim(kwlines(i)))
!       write(minptmpf,'(a)') adjustl(trim(kwlines(i)))
!     enddo

      do i=1,npara                                    ! ZMAT geom spec part of MINP
        write(minpfile,'(a)',advance='no') adjustl(trim(pname(i)))
        write(minpfile,'(3a)',advance='no') '  ='
        write(minpfile,101) fgopt(i)

        write(minptmpf,'(a)',advance='no') adjustl(trim(pname(i)))
        write(minptmpf,'(3a)',advance='no') '  ='
        if(cgopt(i).eq.'  ') then
          write(minptmpf,101) fgopt(i)
        else
          write(minptmpf,103) fgopt(i),cgopt(i)

        endif
      enddo

      write(minpfile,*)
      write(minptmpf,*)

      check = 0
      do while (check.eq.0)                           ! MINP lines after geom spec
          line = ''
          read(mfooterf,204,iostat=check) line
          stmp = adjustl(trim(line))
          l = len_trim(stmp)
          if(check.eq.0) then
            write(minpfile,'(a)') stmp(1:l)
            write(minptmpf,'(a)') stmp(1:l)
          endif
      enddo

!     do i=ilb+1,ilb+ila-1                               ! lines after geom spec
!       write(minpfile,'(a)') adjustl(trim(kwlines(i)))
!       write(minptmpf,'(a)') adjustl(trim(kwlines(i)))
!     enddo

      close(mheaderf)
      close(mfooterf)
      close(minpfile)
      close(minptmpf)

100   format(1000i5)
101   format(1000f22.12)
102   format(i5,a2)
103   format(f22.12,a2)
204   format(a512)
! end update MINP 
      return
      end subroutine update_MINP_ZMAT
!     ******************************************************************

!     ******************************************************************
      subroutine update_MINP_Cart(para,au)
!     ******************************************************************
!     * it updates the MINP file (Cartesian version)
!     ******************************************************************
      implicit none
      double precision, dimension(:), intent(inout) :: para
      double precision, intent(in) :: au ! to backscale the coordinates
      integer :: ipara,check,i,j,k,l
      character(len=512) :: line,stmp

!     do i=1,nopt ! test the parameters to be optimized
!       write(*,*) i, para(i)
!     enddo

      ipara=0
      do i=1,npara ! total parameters in Cartesian, 3*natoms
        if(igopt(i).gt.0) then
          ipara=ipara+1
          fgopt(i)=para(ipara)
        endif
      enddo

!     write(*,*) 'npara: ',npara

      open(minpfile,file='MINP',status='old')
      open(minptmpf,file='MINP.tmp',status='unknown')
      open(mheaderf,file='MINP.header',status='old')
      open(mfooterf,file='MINP.footer',status='old')

      check = 0                                       ! MINP lines before geom spec
      do while (check.eq.0)
          line = ''
          read(mheaderf,204,iostat=check) line
          stmp = adjustl(trim(line))
          l = len_trim(stmp)
          if(check.eq.0) then
            write(minpfile,'(a)') stmp(1:l)
            write(minptmpf,'(a)') stmp(1:l)
          endif
      enddo

      do i=1,npara,3                                  ! MINP geom spec
        k = (i-1)/3 + 1
        write(minpfile,'(10a)',advance='no') adjustl(trim(pname(k)))
        write(minptmpf,'(10a)',advance='no') adjustl(trim(pname(k)))

        do j=0,2
          write(minpfile,101,advance='no') fgopt(i+j)*1.d0/au

          if(cgopt(i+j).eq.'  ') then
            write(minptmpf,101,advance='no') fgopt(i+j)*1.d0/au
          else
            write(minptmpf,103,advance='no') 
     $                             fgopt(i+j)*1.d0/au,cgopt(i+j)
          endif
        enddo

        write(minpfile,'(a)') ''
        write(minptmpf,'(a)') ''

      enddo

      check = 0
      do while (check.eq.0)                           ! MINP lines after geom spec
          line = ''
          read(mfooterf,204,iostat=check) line
          stmp = adjustl(trim(line))
          l = len_trim(stmp)
          if(check.eq.0) then
            write(minpfile,'(a)') stmp(1:l)
            write(minptmpf,'(a)') stmp(1:l)
          endif
      enddo

      close(mheaderf)
      close(mfooterf)
      close(minpfile)
      close(minptmpf)

100   format(1000i5)
101   format(1000f22.12)
102   format(i5,a2)
103   format(f22.12,a2)
204   format(a512)

! end update MINP 
      return
      end subroutine update_MINP_Cart
!     ******************************************************************


************************************************************************
C
      end subroutine geomopt
C
************************************************************************



************************************************************************
